#!/usr/local/bin/perl5 -w

# This program will import a number of files into a particular collection

BEGIN {
    die "GSDLHOME not set\n" unless defined $ENV{'GSDLHOME'};
    die "GSDLOS not set\n" unless defined $ENV{'GSDLOS'};
    unshift (@INC, "$ENV{'GSDLHOME'}/perllib");
    unshift (@INC, "$ENV{'GSDLHOME'}/perllib/plugins");
}

use arcinfo;
use colcfg;
use plugin;
use docsave;


sub print_usage {
    print STDERR "\n  usage: $0 collection-name\n";
}


&main ();

sub main {
    # get and check the collection name
    ($collection) = @ARGV;
    if (!defined($collection)) {
	&print_usage();
	die "\n";
    }
    
    if ($collection eq "modelcol") {
	print STDERR "You can't import documents into modelcol.\n";
	die "\n";
    }
    
    # get the list of plugins for this collection
    @plugins = ("HTMLPlug", "TXTPlug", "RecPlug");
    if (-e "$ENV{'GSDLHOME'}/collect/$collection/etc/collect.cfg") {
	$collectcfg = &colcfg::read_collect_cfg ("$ENV{'GSDLHOME'}/collect/$collection/etc/collect.cfg");
	if (defined $collectcfg->{'plugins'}) {
	    @plugins = @{$collectcfg->{'plugins'}};
	}
    } else {
	print STDERR "WARNING - couldn't find collection configuration file\n";
	print STDERR "          using default plugins\n";
    }
    

    # load all the plugins
    $pluginfo = &plugin::load_plugins ($collection, \@plugins);
    if (scalar(@$pluginfo) == 0) {
	print STDERR "No plugins were loaded.\n";
	die "\n";
    }
    
    # read the archive information file
    $archive_info_filename = "$ENV{'GSDLHOME'}/collect/$collection/archives/archives.inf";
    $archive_info = new arcinfo ();
    $archive_info->load_info ($archive_info_filename);

    # create a docsave object to process the documents
    $processor = new docsave ($collection, $archive_info);

    # process the import directory
    &plugin::read ($pluginfo, "$ENV{'GSDLHOME'}/collect/$collection/import/", 
		   "", {}, $processor);
    
    # write out the archive information file
    $archive_info->save_info($archive_info_filename);
}
