#include <stdlib.h>

#ifdef __GNUG__
#  include <iostream.h>
#  include <fstream.h>
#else
#  ifndef USE_OBJECTSPACE
#    include <iostream>
#    include <fstream>
#  else
#    include <ospace\std\iostream>
#    include <ospace\std\fstream>
#  endif
#endif

#include "cgiwrap.h"
#include "cfgread.h"
#include "fileutil.h"

#ifdef USE_FASTCGI
#include "fcgiapp.h"
#endif

// Note: site.h would not be needed if we could
// dynamically find out gsdlhome.
#include "site.h"


#ifdef USE_FASTCGI
// used to output the text from libinterface
class fcgistreambuf : public streambuf {
public:
  fcgistreambuf ();
  int sync ();
  int overflow (int ch);
  int underflow () {return EOF;}
  
  void fcgisbreset() {fcgx_stream = NULL; other_ostream = NULL;};
  void set_fcgx_stream(FCGX_Stream *newone) {fcgx_stream=newone;};
  void set_other_ostream(ostream *newone) {other_ostream=newone;};
  
private:
  FCGX_Stream *fcgx_stream;
  ostream *other_ostream;
};

fcgistreambuf::fcgistreambuf() {
  fcgisbreset();
  if (base() == ebuf()) allocate();
  setp (base(), ebuf());
};

int fcgistreambuf::sync () {
  if ((fcgx_stream != NULL) &&
      (FCGX_PutStr (pbase(), out_waiting(), fcgx_stream) < 0)) {
    fcgx_stream = NULL;
  }

  if (other_ostream != NULL) {
    char *thepbase=pbase();
    for (int i=0;i<out_waiting();i++) (*other_ostream).put(thepbase[i]);
  }
  
  setp (pbase(), epptr());
  
  return 0;
}

int fcgistreambuf::overflow (int ch) {
  if (sync () == EOF) return EOF;
  if (ch != EOF) sputc (ch);
  return 0;
}

#endif



static int site_cfg_read (libinterface *lib, const text_t &filename,
			  int &maxrequests) {
  text_tarray cfgline;
  char *cstr = filename.getcstr();
  ifstream confin (cstr);
  delete cstr;

  maxrequests = 10000;

  if (confin) {
    while (read_cfg_line(confin, cfgline) >= 0) {
      if (cfgline.size () >= 2) {
	if (cfgline[0] == "gsdlhome") lib->setgsdlhome(cfgline[1]);
	else if (cfgline[0] == "gdbmdir") lib->setgdbmdir(cfgline[1]);
	else if (cfgline[0] == "collectdir") lib->setcollectdir(cfgline[1]);
	else if (cfgline[0] == "httpprefix") lib->sethttpprefix(cfgline[1]);
	else if (cfgline[0] == "gwcgi") lib->setgwcgi(cfgline[1]);
	else if (cfgline[0] == "maxrequests") {
	  maxrequests = cfgline[1].getint();
	  if (maxrequests < 1) maxrequests = 1;
	}
      }
    }
    confin.close ();
    return 1;
  }
  return 0;
}

// cgiwrap does everything necessary to output a page
// using the cgi (or fastcgi) protocol.
int cgiwrap (libinterface *lib) {
#ifdef USE_FASTCGI
  fcgistreambuf outbuf;
#endif

  // init stuff

  // set defaults
  int maxrequests = 10000;
  lib->setgsdlhome (GSDLHOME);
  lib->sethttpprefix("http://127.0.0.1");
  lib->setgwcgi("http://127.0.0.1/cgi-bin/gw");

  // read in the site configuration file
  // try in the collection/etc directory, and then GSDLHOME/etc.
  text_t filename = filename_cat (GSDLHOME, "collect");
  filename = filename_cat (filename, lib->get_collection_name());
  filename = filename_cat (filename, "etc");
  filename = filename_cat (filename, "site.cfg");
  if (!site_cfg_read (lib, filename, maxrequests)) {
    filename = filename_cat (GSDLHOME, "etc");
    filename = filename_cat (GSDLHOME, "site.cfg");
    site_cfg_read (lib, filename, maxrequests);
  }

  // initialise the library software
  ofstream initout (GSDLHOME "/etc/initout.txt");
  lib->init(initout);
  initout.close();
  
  // find out whether this is being run as a cgi-script
  // or a fastcgi script
  int numrequests = 0;
#ifdef USE_FASTCGI
  int isfastcgi = !FCGX_IsCGI();
  FCGX_Stream *fcgiin, *fcgiout, *fcgierr;
  FCGX_ParamArray fcgienvp;
#else
  int isfastcgi = 0;
#endif

  // get the query string if it is not being run as a fastcgi
  // script
  text_t argstr = "";
  char *aURIStr;
  if (!isfastcgi) {
    aURIStr = getenv("QUERY_STRING");
    if (aURIStr != NULL) argstr = aURIStr;
    else {
      char cinURIStr[1024];
      cin >> cinURIStr;
      argstr = cinURIStr;
    }
    maxrequests = 1;
  }

  // Page-request loop. If this is not being run as a fastcgi
  // process then only one request will be processed and then
  // the process will exit.
  while (numrequests < maxrequests) {
#ifdef USE_FASTCGI
    if (isfastcgi) {
      if (FCGX_Accept(&fcgiin, &fcgiout, &fcgierr, &fcgienvp) < 0) break;
      aURIStr = FCGX_GetParam("QUERY_STRING", fcgienvp);
      if (aURIStr != NULL) argstr = aURIStr;
      else argstr = "";
    }
#endif

    // get output streams ready
#ifdef USE_FASTCGI
    outbuf.fcgisbreset ();
    if (isfastcgi) outbuf.set_fcgx_stream (fcgiout);
    else outbuf.set_other_ostream (&cout);
    ostream pageout (&outbuf);
#else
#define pageout cout
#endif
    pageout << "content-type: text/html\n\n";
    ofstream errout (GSDLHOME "/etc/errout.txt");
    cerr = errout;
    
    // get the query page
    lib->getpage (argstr, pageout, errout);
    
    // finish with the output streams
    pageout << "\n";
    pageout << flush;
    errout.close();
#ifdef USE_FASTCGI
    if (isfastcgi) FCGX_Finish();
#endif

    numrequests++;
  }

  return 1;
}

