#ifndef GDBMCLASS_H
#define GDBMCLASS_H

#ifdef __WIN32__
#include "autoconf.h"
#include "systems.h"
#include "gdbmconst.h"
#include "gdbm.h"

#else
#include <gdbm.h>
#endif


class gdbm_info {
public:
  int docnum;            // 'd'
  text_t title;          // 't'
  text_t parent;         // 'p'
  text_t classification; // 'x'
  text_t contents;       // 'c'
  text_t jobnum;         // 'j'
  text_t OID;            // 'o'
  text_t author;         // 'a'
  text_t source;         // 's'
  text_t date;           // 'i'

  void clear ();
};


class gdbmclass {
public:
  gdbmclass() {gdbmfile = NULL; logout = &cerr;};
  ~gdbmclass() {};
  
  // returns 0 if failed, 1 if opened
  int opendatabase (const text_t &filename);
  void closedatabase ();

  // returns 0 on success, -1 on failure
  int getinfo (text_t key, gdbm_info &info);
  void setlogout (ostream *thelogout) {logout = thelogout;}
  int exists (text_t key);

protected:
  text_t openfile;
  GDBM_FILE gdbmfile;
  ostream *logout;
  
  int getkeydata (text_t key, text_t &data);
  int getinfoline (text_t::iterator &here, text_t::iterator end, 
		   text_t &key, text_t &value);
};


// a few useful functions

int is_top_level (const text_t &targetdoc);
int is_book (const text_t &targetdoc);
void get_book_top (const text_t &targetdoc, text_t &book_top);
void get_book (const text_t &targetdoc, text_t &book);
void get_parent_section (text_t &section);
void get_parent_section (text_t &parentsection, text_t &childsection);
int count_dots (text_t::const_iterator first, text_t::const_iterator last);
int count_dots (const text_t &targetdoc);
int is_section_top (const text_t &targetdoc);
void remove_tags (text_t &text);

void seperate_parts (const text_t &targetdoc, gdbmclass &gdbm, const text_t &collection, 
		     text_t &classification, text_t &booksection);
void split_targetdoc (const text_t &targetdoc, text_t &classification, 
		     text_t &booksection);
void splitstring (const text_t &string, vector<text_t> &array);
void get_parents (const text_t &targetdoc, vector<text_t> &parents);
void get_siblings (const text_t &classification, const text_t &booksection, 
		   gdbmclass &gdbm, const text_t &collection, 
		   vector<text_t> &siblings);
int are_same_chapter(text_t section1, text_t section2);
void get_first_section(const text_t &instring, text_t &returnstring);
int is_number (text_t &text);
text_t get_section_str(const text_t &str);
void alphabetize_string_english (text_t &text);
void alphabetize_string_name (text_t &text);
char ** string_add (char **array, int *len, char *str);
void string_sort (char **array, int len);
static int compare_str (const void *e1, const void *e2);
void string_free(char **array, int len);
text_t format_date (const text_t &date);
void format_month (text_t &month);
void  format_day(text_t &day);
#endif
