#ifndef MGSEARCH_H
#define MGSEARCH_H


#include "text_t.h"
#include "querycache.h"

#define RESULTCACHESIZE 10
#define MAXNUMDOCS      500
#define MAXNUMTERMS     100

// a few useful functions for dealing with dir indexes (indexes as they
// appear within the index directory) and real indexes (indexes as they
// appear within the collect.cfg file) and mappings between them.
void getrealdirindex (const text_t &map, text_t &realindex, text_t &dirindex);
bool isdirindex (const text_tarray &indexmap, const text_t &dirindex);
bool isrealindex (const text_tarray &indexmap, const text_t &realindex);
text_t dir2realindex (const text_tarray &indexmap, const text_t &dirindex);
text_t real2dirindex (const text_tarray &indexmap, const text_t &realindex);
text_t real2macroindex (const text_t &realindex);
bool isdoclevelindex (const text_t &realindex);
text_t getdoclevelindex (const text_tarray &indexmap);

class mgsearchclass {
public:
  mgsearchclass ();
  virtual ~mgsearchclass ();
  
  // the index directory must be set before any searching
  // is done
  void setcollectdir (const text_t &thecollectdir);
  
  // the search results are returned in queryresults
  // search returns 'true' if it was able to do a search
  bool search(const queryparamclass &queryparams, 
	      queryresultsclass &queryresults);
  
  // the document text for 'docnum' is placed in 'output'
  // docTargetDocument returns 'true' if it was able to
  // try to get a document
  // collection is needed to see if an index from the 
  // collection is loaded. If no index has been loaded
  // defaultindex is needed to load one
  bool docTargetDocument(const text_t &defaultindex, 
			 const text_t &collection,
			 int docnum,
			 text_t &output);
  
protected:
  querycache *cache;
  text_t collectdir; // the collection directory

  void setsearchmode (const queryparamclass &queryparams);
  void submitquery (const queryparamclass &queryparams);
  void getresults (queryresultsclass &queryresults);

  virtual void extractquoted (text_t &ttquerystring, text_t &ttquotedquery);
  virtual void filterquery (text_t &ttquerystring);
};

#endif

