/**********************************************************************
 *
 * receptionist.h -- a web interface for the gsdl
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * PUT COPYRIGHT NOTICE HERE
 *
 * $Id: receptionist.h 108 1999-01-08 08:41:38Z rjmcnab $
 *
 *********************************************************************/


#ifndef RECEPTIONIST_H
#define RECEPTIONIST_H

#include "text_t.h"
/*
#include "queryinfo.h"
#include "display.h"
#include "gsdlunicode.h"*/
#include "cgiargs.h"
#include "gsdlconf.h"

#ifndef MACROPRECEDENCE
#define MACROPRECEDENCE "style,collection,queryversion,version,language"
#endif


class receptionist {
public:
  receptionist ();
  virtual ~receptionist() {}

  // this version of set_gsdlhome should be used if the receptionist
  // is being run for multiple collections ("general" mode).
  void set_gsdlhome (const text_t &thegsdlhome);

  // this version of set_gsdlhome should be used if the receptionist
  // is being run for a single collection ("collection specific" mode).
  void set_gsdlhome (const text_t &thegsdlhome, const text_t &collection);

  // sets the http address of the gateway cgi program (ie. the program
  // that contains this receptionist).
  void set_gwcgi (const text_t &thegwcgi);

  // init should be called after setgsdhome has been called.
  // It returns true on success and false on failure. If false is
  // returned getpage should not be called (without producing
  // meaningless output), instead an error page should be
  // produced by the calling code.
  bool init (ostream &logout);
  
  // There are two ways to produce a page. You can either call parse_cgi_args,
  // get_cgihead_info, and produce_content or you can just call produce_cgi_page
  // which will be satisfactory in most cases. You might want to call
  // parse_cgi_args, get_cgihead_info, and produce_content when you want to
  // call get_compressed_arg or interface directly with a web server for which
  // the standard header is inappropriate.

  // produce_cgi_page will call parse_cgi_args, get_cgihead_info and 
  // produce_content in the appropriate way to output a cgi header and
  // the page content (if needed).
  void produce_cgi_page (const text_t &argstr, ostream &contentout, 
			 ostream &logout);

  // parse_cgi_args parses cgi arguments into an argument class.
  // This function should be called for each page request.
  void parse_cgi_args (const text_t &argstr, cgiargsclass &args, ostream &logout);

  // response_t is used to inform the calling code what type of
  // cgi header it should produce
  enum response_t {location, content};

  // get_cgihead_info determines the cgi header information for
  // a set of cgi arguments. If response contains location then
  // response_data contains the redirect address. If reponse
  // contains content then reponse_data contains the content-type.
  // Note that images can now be produced by the receptionist.
  void get_cgihead_info (const cgiargsclass &args, response_t &response,
			 text_t &response_data, ostream &logout);

  // produce the page content
  void produce_content (const cgiargsclass &args, ostream &contentout,
			ostream &logout);

  // returns the compressed argument ("e") corresponding to the argument
  // list. This can be used to save preferences between sessions.
  const text_t &get_compressed_arg (const cgiargsclass &args);


protected:

  /*
  cfg_info_t cfg_info;

  text_t gsdlhome;
  text_t gdbmdir;
  text_t collectdir;

  text_t books;
  text_t hrefs;
  text_t titles;
  vector<text_t> bookarr;
  vector<text_t> hrefarr;
  vector<text_t> titlearr;
  int numbooks;

  text_t text_default_index;
  text_t default_index;

  text_t httpprefix;
  text_t gwcgi;
  text_t quotedstring;

  text_t lastcomparg;

  gdbmclass gdbm;
  outconvertclass logconvert;
  displayclass disp;

  inconvertclass asciiinconvert;
  utf8inconvertclass utf8inconvert;
  mapinconvertclass gbinconvert;
  outconvertclass asciioutconvert;
  utf8outconvertclass utf8outconvert;
  mapoutconvertclass gboutconvert;


  browseclass *browse;
  mgsearchclass search;
  
  int cfg_read (const text_t &filename);

  virtual void get_arg_config (text_t &argconfigstr);
  text_t::iterator get_next_config_arg (text_t::iterator first, text_t::iterator last,
					text_t &argname, text_t &defaultvalue, 
					bool &longarg);

  virtual text_t get_compressed_args (cgiargsclass &args);
  virtual void expand_compressed_args (cgiargsclass &args);
  virtual void add_default_args (cgiargsclass &args);
  virtual void check_args (cgiargsclass &args);
  virtual void args_tounicode (cgiargsclass &args, inconvertclass &inconvert);

  virtual void define_general_macros (cgiargsclass &args, outconvertclass &outconvert, 
				      ostream &logout);
  virtual void prepare_page (cgiargsclass &args, outconvertclass &outconvert,
			     ostream &logout);


  virtual void set_query_params (cgiargsclass &args, queryparamclass &queryparams);
  virtual void define_query_macros (cgiargsclass &args, 
				    queryparamclass &queryparams,
				    queryresultsclass &queryresults, 
				    ostream &logout);
  virtual void docLinks (cgiargsclass &args, queryresultsclass &queryresults, 
			 ostream &logout);

  virtual int query_action (cgiargsclass &args, outconvertclass &outconvert, 
			    ostream &textout, ostream &logout);
  virtual void browse_action (cgiargsclass &args, outconvertclass &outconvert, 
			      ostream &textout, ostream &logout);
  virtual void document_action (cgiargsclass &args, outconvertclass &outconvert, 
				ostream &textout, ostream &logout);
  virtual void page_action (cgiargsclass &args, outconvertclass &outconvert, 
			    ostream &textout, ostream &logout);
  virtual void auxiliary_action (cgiargsclass &args, outconvertclass &outconvert, 
				 ostream &textout, ostream &logout);
  
  virtual void displayresults (cgiargsclass &args, outconvertclass &outconvert, 
			       ostream &textout, ostream &logout,
			       queryresultsclass &queryresults);

  virtual void define_pref_macros (cgiargsclass &args, ostream &logout);
  virtual int do_query(cgiargsclass &args, queryparamclass &queryparams, 
			queryresultsclass &queryresults, ostream &logout);
  virtual void highlighttext(text_tarray &termvars, outconvertclass &outconvert, 
			     ostream &textout, ostream &logout, text_t &text);
  void format_querystring (text_t &querystring);


  virtual void define_collection_macros (cgiargsclass &args, ostream &logout) {}
  virtual bool collection_init(const text_t &collection) {return true;}
  virtual void displaydocsummary (cgiargsclass &args, outconvertclass &outconvert, 
				  ostream &textout, ostream &logout, 
				  queryresultsclass &queryresults,
				  int resultnum) {}

  */
};



#endif
